@@ -207,7 +207,7 @@ trait Printers extends api.Printers { self: SymbolTable =>
     )
 
     def printFlags(flags: Long, privateWithin: String) = {
-      val mask: Long = if (settings.debug) -1L else PrintableFlags
+      val mask: Long = if (settings.isDebug) -1L else PrintableFlags
       val s = flagsToString(flags & mask, privateWithin)
       if (s != "") print(s + " ")
     }
@@ -320,7 +320,7 @@ trait Printers extends api.Printers { self: SymbolTable =>
       if (qual.nonEmpty || (checkSymbol && tree.symbol != NoSymbol)) print(resultName + ".")
       print("super")
       if (mix.nonEmpty) print(s"[$mix]")
-      else if (settings.debug) tree.tpe match {
+      else if (settings.isDebug) tree.tpe match {
         case st: SuperType => print(s"[${st.supertpe}]")
         case tp: Type => print(s"[$tp]")
         case _ =>
@@ -479,7 +479,7 @@ trait Printers extends api.Printers { self: SymbolTable =>
         case th @ This(qual) =>
           printThis(th, symName(tree, qual))
 
-        case Select(qual: New, name) if !settings.debug =>
+        case Select(qual: New, name) if !settings.isDebug =>
           print(qual)
 
         case Select(qualifier, name) =>
@@ -781,26 +781,30 @@ trait Printers extends api.Printers { self: SymbolTable =>
             print("class ", printedName(name))
             printTypeParams(tparams)
 
-            val build.SyntacticClassDef(_, _, _, ctorMods, vparamss, earlyDefs, parents, selfType, body) = cl: @unchecked
+            cl match {
+              case build.SyntacticClassDef(_, _, _, ctorMods, vparamss, earlyDefs, parents, selfType, body) =>
+                // constructor's modifier
+                if (ctorMods.hasFlag(AccessFlags) || ctorMods.hasAccessBoundary) {
+                  print(" ")
+                  printModifiers(ctorMods, primaryCtorParam = false)
+                }
 
-            // constructor's modifier
-            if (ctorMods.hasFlag(AccessFlags) || ctorMods.hasAccessBoundary) {
-              print(" ")
-              printModifiers(ctorMods, primaryCtorParam = false)
+                def printConstrParams(ts: List[ValDef]): Unit = {
+                  parenthesize() {
+                    printImplicitInParamsList(ts)
+                    printSeq(ts)(printVParam(_, primaryCtorParam = true))(print(", "))
+                  }
+                }
+                // constructor's params processing (don't print single empty constructor param list)
+                vparamss match {
+                  case Nil | List(Nil) if !mods.isCase && !ctorMods.hasFlag(AccessFlags) =>
+                  case _ => vparamss foreach printConstrParams
+                }
+                parents
+              case _ =>
+                // Can get here with erroneous code, like `{@deprecatedName `
+                Nil
             }
-
-            def printConstrParams(ts: List[ValDef]): Unit = {
-              parenthesize() {
-                printImplicitInParamsList(ts)
-                printSeq(ts)(printVParam(_, primaryCtorParam = true))(print(", "))
-              }
-            }
-            // constructor's params processing (don't print single empty constructor param list)
-            vparamss match {
-              case Nil | List(Nil) if !mods.isCase && !ctorMods.hasFlag(AccessFlags) =>
-              case _ => vparamss foreach printConstrParams
-            }
-            parents
           }
 
           // get trees without default classes and traits (when they are last)
