@@ -14,22 +14,23 @@ package scala.tools.nsc.interpreter
 package shell
 
 trait Completion {
-  def complete(buffer: String, cursor: Int): CompletionResult
+  final def complete(buffer: String, cursor: Int): CompletionResult = complete(buffer, cursor, filter = true)
+  def complete(buffer: String, cursor: Int, filter: Boolean): CompletionResult
 }
 object NoCompletion extends Completion {
-  def complete(buffer: String, cursor: Int) = NoCompletions
+  def complete(buffer: String, cursor: Int, filter: Boolean) = NoCompletions
 }
 
-case class CompletionResult(line: String, cursor: Int, candidates: List[CompletionCandidate]) {
+case class CompletionResult(line: String, cursor: Int, candidates: List[CompletionCandidate], typeAtCursor: String = "", typedTree: String = "") {
   final def orElse(other: => CompletionResult): CompletionResult =
     if (candidates.nonEmpty) this else other
 }
 object CompletionResult {
   val empty: CompletionResult = NoCompletions
 }
-object NoCompletions extends CompletionResult("", -1, Nil)
+object NoCompletions extends CompletionResult("", -1, Nil, "", "")
 
 case class MultiCompletion(underlying: Completion*) extends Completion {
-  override def complete(buffer: String, cursor: Int) =
-    underlying.foldLeft(CompletionResult.empty)((r, c) => r.orElse(c.complete(buffer, cursor)))
+  override def complete(buffer: String, cursor: Int, filter: Boolean) =
+    underlying.foldLeft(CompletionResult.empty)((r,c) => r.orElse(c.complete(buffer, cursor, filter)))
 }
