@@ -18,6 +18,7 @@ import java.lang.reflect.InvocationTargetException
 import java.nio.charset.Charset
 import java.nio.file.{Files, StandardOpenOption}
 
+import scala.annotation.nowarn
 import scala.collection.mutable.ListBuffer
 import scala.concurrent.duration.Duration
 import scala.reflect.internal.FatalError
@@ -258,7 +259,10 @@ class Runner(val testInfo: TestInfo, val suiteRunner: AbstractRunner) {
 
       pushTranscript(s"<in process execution of $testIdent> > ${logFile.getName}")
 
-      TrapExit(() => run()) match {
+      @nowarn("cat=deprecation")  // JDK 17 deprecates SecurityManager, so TrapExit is deprecated too
+      val trapExit = TrapExit
+
+      trapExit(() => run()) match {
         case Left((status, throwable)) if status != 0 =>
           genFail("non-zero exit code")
         case _ =>
